import os
import os.path
import shutil
import string
import logging
import kraken.tools.Utils as Utils
import kraken.Constants
from kraken.entrytypes.Entry import Entry, loadKraEntry, getEntryRdfFile
from kraken.entrytypes.MetaData import MetaData, loadMetaData, generateCache, updateCache
from kraken.tools.Utils import cleanExit
import kraken.tools.TextUi as TextUi

def description(system, runtime):
    return """
    "Creates a new entry either from a local file or or a webpage.
    In the future, creation from scratch and hollow entries will also be supported\n"
    """
#
#  Creates an entry
# 
def execute(system, runtime):    
    sources = interactiveCreate(system, runtime)
    metadata = loadMetaData(system, runtime)        
    temp = sources[0]    
    if temp == "":
        entry = createNote(system, runtime)
    elif (temp[0:7] == "http://"):
        if TextUi.confirm("This is a webpage. Create only a link?"):
            entry = createWebpageLink(system, runtime, sources)
        else:    
            if TextUi.confirm("Download the full webpage (with images, icons etc?)"):
               entry = createEntryForFullWebpage(system, runtime, sources)
            else:
               entry = createEntryForSingleWebpage(system, runtime, sources)
    else: # Local files 
        entry = createEntryForSingleFile(system, runtime, sources)
    if entry == None:
        return
    entry.fillinEntryFromRuntime(runtime, system)
    # small trick to allow editing while creating
    x = entry.parentDirName
    entry.parentDirName = system.getProperty("tempDir")
    entry.edit(metadata, isAnnotation = 0)
    entry.parentDirName = x 
    
    entry.writeRDF(os.path.join(system.getProperty("tempDir"), runtime.entryDir))
    # print entry
    Utils.moveEntryToDestinationDir(system, runtime)
    print "Kraken entry " + runtime.entryDir + " successfully created. "
    # refreshing the metadata 
    metadata.addEntry(entry)
    updateCache(system, runtime, metadata)

    
def interactiveCreate(system, runtime):
    '''
       Interactive create
    '''    
    print """
      Enter the source(s) of a new entry. 
      Type in a webpage, a local file, or simply press <ENTER> 
    to create an entry without an external source.
      End your lines with a comma (,) to enter more than one source.
    """
    value = []
    notDone = 1
    while notDone:
       temp = TextUi.inputString("Source")        
       if temp == "":
          if len(value)==0:
             value = [""]
          return value
       # done / not done
       if temp[-1] == ",":
           temp = temp[0:-1]
           notDone = 1
       else: 
           notDone = 0
       # interpret as webpage
       if temp[0:7] == "http://":
          print "WEBPAGE: " + temp
          value.append(temp)
          continue
       # interpret as file   
       filename = os.path.expanduser(temp)
       filename = os.path.abspath(filename)
       if os.path.exists(filename):
          print "FILE: " + filename
          value.append(filename)
          continue
       else:
          print temp + " is not a valid file!!!"    
    return value
    

"""    
  Create the entry for a single file from the local filesystem
"""
def createEntryForSingleFile(system, runtime, sources):
    print "Creating an entry based on local files"
    print "Checking if the files exist..."
    localfiles = []
    for filename in sources:
       filename = os.path.expanduser(filename)
       filename = os.path.abspath(filename)
       if not os.path.exists(filename):
           logging.critical(filename + " does not exist")
           return None
       localfiles.append(filename)    
    Utils.createEntry(system, runtime)
    print "Copying the files..."
    for filename in localfiles:    
       shutil.copy2(filename, os.path.join(system.getProperty("tempDir"), runtime.entryDir))
    entry = Entry()
    entry.addProperty("originalURIs", localfiles)
    entry.localFileNames = []
    for filename in localfiles:
       entry.localFileNames.append(os.path.split(filename)[1])
    print "Kraken entry " + runtime.entryDir + " successfully created"
    return entry

def createNote(system, runtime):
    Utils.createEntry(system, runtime)
    targetdir = os.path.join(system.getProperty("tempDir"), runtime.entryDir)
    # now, create the kraken file
    entry = Entry()
    return entry;
    

def createWebpageLink(system, runtime, sources):
    Utils.createEntry(system, runtime)
    targetdir = os.path.join(system.getProperty("tempDir"), runtime.entryDir)
    # now, create the kraken file
    entry = Entry()
    entry.addProperty("originalURIs", sources)
    return entry;
    
  
# ***********************************************************
#  Create a kraken entry for a single webpage
# ***********************************************************
def createEntryForSingleWebpage(system, runtime, sources):
    Utils.createEntry(system, runtime)
    targetdir = os.path.join(system.getProperty("tempDir"), runtime.entryDir)
    if len(sources) > 1:
        logging.critical("You should enter a single web address!")
        return
    command = (kraken.Constants.COMMAND_DOWNLOAD_SINGLE_FILE + '"' + 
           sources[0] + '"' +
          ' --directory-prefix="' + targetdir + '"')
    logging.info("Executing: " + command)
    os.system(command)
    # now, create the kraken file
    entry = Entry()
    entry.addProperty("originalURIs", sources)
    # get the http filename, simply picking the first file in the directory    
    for x in os.listdir(targetdir):
       if not os.path.isdir(x):
           entry.addProperty("localFileNames", [x])
    return entry;

# **********************************************************
#  Create the kraken entry for a full webpage
# ***********************************************************
def createEntryForFullWebpage(system, runtime, sources):
    if os.name == "nt":
        # targetdir = "\"" + string.replace(targetdir,"/","\\\\") + "\""
        print "This command does not work correctly in Windows, because of some file"
        print "name mapping issues. It is a wget problem, but until I find a solution"
        print "I have disabled it"
        os._exit(0)       
    Utils.createEntry(system, runtime)
    targetdir = os.path.join(system.getProperty("tempDir"), runtime.entryDir)
    print targetdir      
    command = (kraken.Constants.COMMAND_DOWNLOAD_FULLPAGE + sources[0] +
      " --directory-prefix="+targetdir)
    print command
    os.system(command)
    # now, create the cocoon file
    entry = Entry()
    entry.addProperty("originalURIs", sources)
    temp = sources[0][7:] # got rid of http://
    # get the http filename, simply picking the first file in the directory
    resultFile = os.path.join(system.getProperty("tempDir"), runtime.entryDir, temp)
    print("Resultfile: " + resultFile)
    if os.path.isdir(resultFile):    
       listdir = os.listdir(resultFile)
       for priv in ("index.html", "index.htm", os.path.split(runtime.inputFile)[1]):
          if priv in listdir:
             entry.addProperty("localFileNames", [os.path.join(temp,priv)]) 
             logging.info("Filename is:" + entry.localFileNames[0])
          else:
             print("Could not determine the local file name")
    elif os.path.exists(resultFile):
        entry.addProperty("localFileNames", [temp])
    else:
        logging.error("Can not identify the main file"+temp)
    return entry

